#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Sep 25 12:03:18 2020

@author: Jose Rosas

This script calculates the pIC50 values for cruzain inhibition
from molecular structures. Details on modeling and validation are
found in the paper:
    Rosas-Jimenez, J.G., Garcia-Revilla, M.A., Madariaga-Mazon, A.,
    Martinez-Mayorga, K. Predictive global models of
    cruzain inhibitors with large chemical coverage. ACS Omega. 2021

Basically, it calculates mordred descriptors for molecules and
apply a KNN and RF pre-trained model to predict the activity values.

Usage: qsar_cruzain_predictor.py [-h] [-i input.sdf] [-o output.csv]
Arguments:
    -h, --help Help message
    -i,        SDFile with 2D structures. Default is input.sdf
    -o,        Name of the csv to write results. Default is output.csv
"""
import sys
import os
import pickle
import argparse
import numpy as np
import pandas as pd
from rdkit import Chem
from rdkit.Chem import AllChem
from mordred import Calculator
from mordred.AcidBase import AcidicGroupCount
from mordred.AtomCount import AtomCount
from mordred.Autocorrelation import MATS, GATS, AATSC
from mordred.BCUT import BCUT
from mordred.EState import AtomTypeEState
from mordred.InformationContent import ComplementaryIC
from mordred.MoeType import PEOE_VSA, SlogP_VSA
from mordred.TopologicalCharge import TopologicalCharge
from utilities import ADSimilarity, ADLeverage

def main(input_name, output_name):
    #If no input file is provided, print help message and stops the program
    if input_name is True:
        print('Input sdf file was not provided, use -h flag to show usage \
    or consult the README file for instructions')
        sys.exit(1)
    
    #Using RDKit to read the molecules in the sdf file. It also recalculates
    #2D coordinates
    print('Rebuilding molecules...')
    mols = [mol for mol in Chem.SDMolSupplier(input_name)]
    molsh = [Chem.AddHs(mol) for mol in mols]
    mol_ids = np.array([mol.GetProp('_Name') for mol in mols])
    for mol in molsh: AllChem.Compute2DCoords(mol)
    
    #Loads the pickle objects with the trained models
    #with open('./utils/KNN_CRUZ.pickle', 'rb') as reader:
    #    model_knn = pickle.load(reader)
    with open(os.path.join('utils', 'KNN_CRUZ.pickle'), 'rb') as reader:
        model_knn = pickle.load(reader)
    
    #with open('./utils/RF_CRUZ.pickle', 'rb') as reader:
    #    model_rf = pickle.load(reader)
    with open(os.path.join('utils', 'RF_CRUZ.pickle'), 'rb') as reader:
        model_rf = pickle.load(reader)
    
    #Loads pickle objects with information about the training set
    #for applicability domain calculation
    #with open('./utils/AD_KNN.pickle', 'rb') as reader:
    #    ad_knn = pickle.load(reader)
    with open(os.path.join('utils', 'AD_KNN.pickle'), 'rb') as reader:
        ad_knn = pickle.load(reader)
    
    #with open('./utils/AD_RF.pickle', 'rb') as reader:
    #    ad_rf = pickle.load(reader)
    with open(os.path.join('utils', 'AD_RF.pickle'), 'rb') as reader:
        ad_rf = pickle.load(reader)
        
    #with open('./utils/ADsimilarity.pickle', 'rb') as reader:
    #    ad_sim = pickle.load(reader)
    with open(os.path.join('utils', 'ADsimilarity.pickle'), 'rb') as reader:
        ad_sim = pickle.load(reader)
    
    #Calculation of descriptor for the new molecules using mordred
    print('Calculating descriptors...')
    desc_calculator = Calculator()
    desc_calculator.register(MATS(5,'Z'))
    desc_calculator.register(GATS(3, 'c'))
    desc_calculator.register(GATS(8, 's'))
    desc_calculator.register(BCUT('c', 0))
    desc_calculator.register(AtomTypeEState('count', 'sssCH'))
    desc_calculator.register(ComplementaryIC(0))
    desc_calculator.register(PEOE_VSA(4))
    desc_calculator.register(TopologicalCharge('mean', 4))
    desc_calculator.register(TopologicalCharge('mean', 8))
    desc_calculator.register(AcidicGroupCount())
    desc_calculator.register(AtomCount('F'))
    desc_calculator.register(AATSC(1, 'c'))
    desc_calculator.register(AATSC(1, 'dv'))
    desc_calculator.register(BCUT('d', -1))
    desc_calculator.register(AtomTypeEState('count', 'dsN'))
    desc_calculator.register(AtomTypeEState('count', 'dO'))
    desc_calculator.register(PEOE_VSA(5))
    desc_calculator.register(SlogP_VSA(2))
    
    data = desc_calculator.pandas(molsh)
    X_knn = data.to_numpy()[:,:9]
    X_rf = data.to_numpy()[:,9:]
    
    #Calculates the predicted activity values using the models and then the
    #applicability domain parameters 
    print('Predicting pIC50 values...')
    y_knn = model_knn.predict(X_knn)
    y_rf = model_rf.predict(X_rf)
    
    print('Calculating Applicability Domain...')
    similarity = ADSimilarity(ad_sim, molsh)
    leverage_knn = ADLeverage(ad_knn, X_knn)
    leverage_rf = ADLeverage(ad_rf, X_rf)
    
    similarities = similarity.calculate()
    leverages_knn = leverage_knn.calculate()
    leverages_rf = leverage_rf.calculate()
    
    #Saves the results in a csv file
    print('Writing final results...')
    data['pIC50_KNN'] = pd.Series(y_knn)
    data['pIC50_RF'] = pd.Series(y_rf)
    data['Max Similarity'] = pd.Series(similarities['Max'])
    data['Leverage KNN'] = pd.Series(leverages_knn['Leverage'])
    data['Max Leverage KNN'] = pd.Series(leverages_knn['Limit'])
    data['Leverage RF'] = pd.Series(leverages_rf['Leverage'])
    data['Max Leverage RF'] = pd.Series(leverages_rf['Limit'])
    data['ID'] = pd.Series(mol_ids)
    data.to_csv(output_name, index=False)

if __name__ == '__main__':
    #Parse arguments from the command line
    parser = argparse.ArgumentParser(description='QSAR cruzain inhibition predictor')
    parser.add_argument('-i', nargs=1, default=[True],
                        help='SDFile with 2D structures.')
    parser.add_argument('-o', nargs=1, default=['output.csv'],
                        help='Name of the csv to write results. Default is output.csv')
    args = parser.parse_args()
    input_name = args.i[0]
    output_name = args.o[0]
    
    main(input_name, output_name)
