#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Sep 25 14:39:38 2020

@author: Jose Rosas

Definition of some classes useful in applicability domain calculation
to be used with the qsar_cruzain_predictor.py script.
"""
import numpy as np
from rdkit import DataStructs
from rdkit.Chem import MACCSkeys

class ADSimilarity(object):
    def __init__(self, training_molecules, test_molecules):
        self.training_molecules = training_molecules
        self.test_molecules = test_molecules
        self.fps_train = None
        self.fps_test = None
        self.similarity_matrix = None
    def __fingerprinting(self):
        self.fps_train = [MACCSkeys.GenMACCSKeys(mol) for mol in self.training_molecules]
        self.fps_test = [MACCSkeys.GenMACCSKeys(mol) for mol in self.test_molecules]
    def __similarity_matrix(self):
        self.similarity_matrix = np.zeros([len(self.fps_test),len(self.fps_train)])
        for i in range(len(self.fps_test)):
            for j in range(len(self.fps_train)):
                self.similarity_matrix[i,j] = DataStructs.FingerprintSimilarity(self.fps_test[i], self.fps_train[j])
    def calculate(self):
        self.__fingerprinting()
        self.__similarity_matrix()
        summary = {'Min':self.similarity_matrix.min(axis=1),
                   'Max':self.similarity_matrix.max(axis=1),
                   'Mean':self.similarity_matrix.mean(axis=1),
                   '1st quartile':np.percentile(self.similarity_matrix, 25, axis=1),
                   '2nd quartile':np.percentile(self.similarity_matrix, 50, axis=1),
                   '3rd quartile':np.percentile(self.similarity_matrix, 75, axis=1)}
        return summary

class ADLeverage(object):
    def __init__(self, training_molecules, test_molecules):
        self.training_molecules = training_molecules
        self.test_molecules = test_molecules
        self.threshold = (3*training_molecules.shape[1]) / training_molecules.shape[0]
    def __hat_matrix(self):
        H = self.test_molecules.dot(np.linalg.inv(self.training_molecules.T.dot(self.training_molecules)).dot(self.test_molecules.T))
        return H
    def calculate(self):
        leverages = np.diagonal(self.__hat_matrix())
        limit = np.repeat(self.threshold, self.test_molecules.shape[0])
        summary = {'Leverage':leverages, 'Limit':limit}
        return summary
